#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netdb.h>
#include <signal.h>
#include <memory.h>
#include <sys/errno.h>
#include <fcntl.h>

#ifdef _HAVE_SELECT_H
#include <sys/select.h>
#endif

#include "player.h"
#include "server.h"

struct connection conns[ARRAYSIZE];

int numconns = 0;
int mastsock;

int main (int argc, char *argv[])
{
int port;
int retval;

	write_PID ();
	clearsocks ();
    set_signals ();

    if (argc > 2) {
        printf("Usage: gcs-server [port]\n");
        printf("If no port is specified, %d is used.\n", PORT);
        exit(1);
    }
    if (argc == 2)
        port = atoi(argv[1]);  /* port is argument one */
    else
        port = PORT;     /* or PORT by default */

	retval = server_init_and_loop (port);
	if (retval == -1)
		serverlog ("EXIT: abnormal exit.  Received -1 from server_init_and_loop()\n");
	else
		serverlog ("EXIT: Normal exit.\n");
	exit(0);
}

int server_init_and_loop (int port)
{
struct          sockaddr_in cli_addr;
int                             alen;
int                           numFDS;
int                           retval;
register int                    curr;
fd_set                       readFDS;
unsigned long int        longaddr, a;
char                      errbuf[80];
struct hostent                 *hent;
char                 tempaddress[40];

#ifndef _NO_BZERO
    bzero ((char *)&cli_addr, sizeof(cli_addr));
#else
	memset ((char *)&cli_addr, '\0', sizeof(cli_addr));
#endif
    cli_addr.sin_family = AF_INET;
    cli_addr.sin_addr.s_addr = INADDR_ANY;
    cli_addr.sin_port = htons((u_short)port);
    mastsock = socket (PF_INET, SOCK_STREAM, 0);
    if (mastsock < 0) {
		perror ("socket");
        return(-1);
    }
    if (bind(mastsock, (struct sockaddr *)&cli_addr, sizeof(cli_addr)) < 0) {
		perror ("bind");
        return(-1);
    }
    if (listen (mastsock, QUEUE_LENGTH) < 0) {
		perror ("listen");
        return(-1);
    }

    retval = setsockopt (mastsock, SOL_SOCKET, SO_REUSEADDR, (char *) 1, 1);
#ifndef _NO_GETDTABLESIZE
    numFDS = getdtablesize();
#else
	numFDS = sysconf (_SC_OPEN_MAX) - 8;
#endif

    while (1) {

        FD_ZERO (&readFDS);
        FD_SET (mastsock, &readFDS);
        for (curr=0; curr < numconns; curr++)
            FD_SET (conns[curr].sock, &readFDS);

        if (select(numFDS, &readFDS, NULL, NULL, NULL) < 0) {
			sprintf (errbuf, "SELECT: %d\n", errno);
            crash(errbuf);
            return (-1);
        }

        if (FD_ISSET(mastsock, &readFDS)) {
            if ((curr = new_conn()) < 0) {
				crash ("FATAL : malloc returned NULL.\n");
				return (1);
			}
			/* numconns now is one greater */
			strcpy (conns[curr].name, "UnConnPlayer");
			strcpy (conns[curr].address, "Connecting");
            alen = sizeof(conns[curr].cli_addr);
            conns[curr].sock = accept(mastsock, 
							(struct sockaddr *)&(conns[curr].cli_addr), &alen);
            if (conns[curr].sock < 0) {
				sprintf (errbuf, "accept: %d\n", errno);
                crash(errbuf);
			} else {
#ifdef O_NONBLOCK
				fcntl (conns[curr].sock, F_SETFL, O_NONBLOCK);
#else
				fcntl (conns[curr].sock, F_SETFL, FNDELAY);
#endif
                spewfile (curr, "opening");
        		longaddr = ntohl(conns[curr].cli_addr.sin_addr.s_addr);
        		hent = gethostbyaddr ((char *) &longaddr, sizeof (longaddr), AF_INET);
        		if (!hent) {
					a = longaddr;
					sprintf (tempaddress, "%d.%d.%d.%d", (a >> 24) & 0xff,
                    (a >> 16) & 0xff, (a >> 8) & 0xff, a & 0xff);
					strcpy (conns[curr].address, tempaddress);
        		} else
					strcpy (conns[curr].address, hent->h_name);
			}
        }
        for (curr=0; curr < numconns; curr++)
            if (FD_ISSET(conns[curr].sock, &readFDS))
                action(curr);
	}
}
